#!/bin/python3
'''
Reese Grimsley 8/11/22

Written for generic execution of ONNX models for testing purposes. 
Read from input file and print model output to terminal
Read in YAML params file for preprocessing

'''

import os, sys, time
from pyexpat import model
from pprint import pprint
import numpy as np
import onnxruntime
import tflite_runtime.interpreter as tflitert_interpreter
import cv2 as cv
import yaml
import argparse


so = onnxruntime.SessionOptions()
ort_logger_level = 0
#so.log_severity_level=ort_logger_level
#onnxruntime.set_default_logger_severity (ort_logger_level)
so.graph_optimization_level = onnxruntime.GraphOptimizationLevel.ORT_ENABLE_ALL  #default is ORT_ENABLE_ALL. Others are ORT_ENABLE_EXTENDED, ORT_ENABLE_BASIC, ORT_DISABLE_ALL


parser = argparse.ArgumentParser()
parser.add_argument('modeldir', type=str, default=None)
parser.add_argument('-t', '--TIDL', default=True, action='store_true')
parser.add_argument('-nt', '--no-TIDL', dest='TIDL', action='store_false')
parser.add_argument('-d', '--debug-level', default=0)
parser.add_argument('-n', '--num-iter', default=25, type=int)
parser.add_argument('-p', '--print-result', default=False, action='store_true')
args = parser.parse_args()
#modeldir = '/opt/model_zoo/ONR-OD-8020-ssd-lite-mobv2-mmdet-coco-512x512'
modeldir = args.modeldir 
#modelfile = os.path.join(modeldir, 'model', 'ssd_mobilenetv2_lite_512x512_20201214_model.onnx')



infile = 'cat-big.jpeg'
#infile = 'ADE_val_00001801.jpg'
#infile = 'car.jpeg'
paramsfile = os.path.join(modeldir,'param.yaml')
params = yaml.safe_load(open(paramsfile, 'r'))

modelfile = os.path.join(modeldir, params['session']['model_path'])
assert os.path.exists(modelfile)

outfile = ''
#outputdir = os.path.join(modeldir, 'artifacts')
outputdir = os.path.join(modeldir, params['session']['artifacts_folder'])

has_TIDL = args.TIDL

runtime = None
if '.onnx' in modelfile: runtime = 'onnx'
elif '.tflite' in modelfile: runtime = 'tflite'
else: raise ValueError('Unable to determine runtime from filename..')

runtime_options = {
    "tidl_platform": "J7",
    "tidl_version": "7.2",
    "tidl_tools_path": "null",
    "artifacts_folder": outputdir,
    "tensor_bits": 8,
    "import": 'yes',
    "debug_level": args.debug_level,
}

def get_benchmark_output(benchmark_dict):
    proc_time = copy_time = 0
    cp_in_time = cp_out_time = 0
    subgraphIds = []
    for stat in benchmark_dict.keys():
        if 'proc_start' in stat:
            subgraphIds.append(stat.replace('ts:subgraph_', '').replace('_proc_start', ''))
    for i in range(len(subgraphIds)):
        proc_time += benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_proc_end'] - benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_proc_start']
        cp_in_time += benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_copy_in_end'] - benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_copy_in_start']
        cp_out_time += benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_copy_out_end'] - benchmark_dict['ts:subgraph_'+str(subgraphIds[i])+'_copy_out_start']
    copy_time = cp_in_time + cp_out_time
    copy_time = copy_time if len(subgraphIds) == 1 else 0
    total_time   = benchmark_dict['ts:run_end'] - benchmark_dict['ts:run_start']
    read_total  = benchmark_dict['ddr:read_end'] - benchmark_dict['ddr:read_start']
    write_total   = benchmark_dict['ddr:write_end'] - benchmark_dict['ddr:write_start']

    total_time = total_time - copy_time

    return total_time/1000000, proc_time/1000000, read_total/1000000, write_total/1000000

def prep_model_onnx(modelfile, sess_options=None):
    if sess_options is None:
        sess_options = onnxruntime.SessionOptions()
    if has_TIDL:
        ep_list = ['TIDLExecutionProvider', 'CPUExecutionProvider']
        ep_options = [runtime_options, {}]
    else:
        ep_list = ['CPUExecutionProvider']
        ep_options = [{}]

    interpreter = onnxruntime.InferenceSession(modelfile, providers=ep_list, provider_options=ep_options, sess_options=sess_options)
    input_details = interpreter.get_inputs()
    
    return interpreter, input_details, sess_options

def prep_model_tflite(modelfile):
    if has_TIDL:
        tidl_delegate = [tflitert_interpreter.load_delegate('libtidl_tfl_delegate.so', runtime_options)]
        interpreter = tflitert_interpreter.Interpreter(modelfile, experimental_delegates=tidl_delegate)
    else:
        interpreter = tflitert_interpreter.Interpreter(model_path=modelfile)
    
    interpreter.allocate_tensors()
    input_details = interpreter.get_input_details()
    output_details = interpreter.get_output_details()
    return interpreter, input_details, output_details

def read_and_preprocess_input(filename, paramsfile):
    input_image = None
    with open(paramsfile) as yaml_file:
        try: 
            model_config = yaml.safe_load(yaml_file)
            # pprint(model_config)
            input_image = cv.imread(infile)
            size = model_config['preprocess']['crop']
            if type(size) == int:
                size=(size,size) #(width,heighti)
            else:
                size = (size[1], size[0]) #width, height; param.yaml formats as height,width
            print(size)
            input_image = cv.resize(input_image, size)
            
            # subtract mean and scale image if this was done during training
            if model_config['session']['input_mean'] is not None:
                input_image = input_image.astype('float32') - model_config['session']['input_mean']
            if model_config['session']['input_scale'] is not None:
                input_image = input_image.astype('float32') * model_config['session']['input_scale']
            #swap channel
            if model_config['preprocess']['data_layout'] == 'NCHW':
                input_image = np.expand_dims(np.transpose(input_image, (2,0,1)), axis=0)
            else:
                input_image = np.expand_dims(input_image, axis=0)
                
        except yaml.YAMLError as e:
            print (e)
        
    return input_image
    

if __name__ == '__main__':
    img = read_and_preprocess_input(infile, paramsfile)
    if runtime == 'onnx':
        model, input_details, sess_options =  prep_model_onnx(modelfile, so)
        i = input_details[0]
        dtype = i.type.split('(')[-1][:-1] #format of type is "tensor($TYPE)", e.g. "tensor(uint8)"
        input_shape = i.shape
        #print(input_shape)
        n,c,h,w = input_shape
    elif runtime == 'tflite':
        interpreter, model_input, model_output = prep_model_tflite(modelfile)
        #print(model_input[0])
        dtype = model_input[0]['dtype']
        input_shape = model_input[0]['shape']
        n,h,w,c = input_shape

    #parse model input details in case data type needs to change
    if dtype in ['uint8', 'int8', 'uint16', 'int16']:
        img = img.astype(dtype)
    elif dtype == 'float' or 'float32' in str(dtype):
        img = img.astype('float32')
    elif 'uint8' in dtype: 
        img = img.astype(np.uint8)
    elif 'int8' in dtype: 
        img = img.astype(np.int8)
    elif 'uint16' in dtype: 
        img = img.astype(np.uint16)
    elif 'int16' in dtype: 
        img = img.astype(np.int16)
    else: 
        print(f'Unknown data type {dtype}')
  
    if c == 1:
        img = img[0,0,...][None,None,...]

    print('running model now...')
    sum_time = 0   
    min_time = 1000000000000
    iterations = args.num_iter 
    import time
    print(img.shape)
    for i in range(iterations):
        if runtime == 'onnx':
            t1=time.time()
            result = model.run(None, {input_details[0].name: img})
            t2=time.time()    
            if args.print_result: print(result)
        elif runtime == 'tflite':
            interpreter.set_tensor(model_input[0]['index'], img)
            t1=time.time()
            interpreter.invoke()
            t2=time.time() 
        t_elapse = t2-t1
        sum_time += t_elapse
        print(t_elapse)
        if t_elapse < min_time: min_time = t_elapse


print('minimum execution time is %.3f ms' % (min_time * 1000))
print('Total time to run %d iterations was %.4f s'% (iterations, sum_time))
if runtime == 'onnx':
    #print(result)
    stats = model.get_TI_benchmark_data()
elif runtime == 'tflite':
    stats = interpreter.get_TI_benchmark_data()

print(get_benchmark_output(stats)) 
print(stats)
